'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

The watcher is actually a connection to crontab. This is what helps to schedule
the watched to check for changes at some frequency, and update the files.


'''

import os
import sys

import sys
import time
import threading
from random import choice

class Spinner:
    spinning = False
    delay = 0.1

    @staticmethod
    def spinning_cursor():
        while 1: 
            for cursor in '|/-\\': yield cursor

    @staticmethod
    def balloons_cursor():
        while 1: 
            for cursor in '. o O @ *': yield cursor

    @staticmethod
    def changing_arrows():
        while 1: 
            for cursor in '<^>v': yield cursor

    def select_generator(self, generator):
        if generator == None:
            generator = choice(['cursor',
                                'arrow',
                                'balloons'])

        return generator

    def __init__(self, delay=None, generator=None):
        generator = self.select_generator(generator)

        if generator == 'cursor':
            self.spinner_generator = self.spinning_cursor()
        elif generator == 'arrow':
            self.spinner_generator = self.changing_arrows()
        elif generator == 'balloons':
            self.spinner_generator = self.balloons_cursor()
            if delay is None: delay = 0.2
        else:
            self.spinner_generator = self.spinning_cursor()

        if delay and float(delay):
            self.delay = delay

    def run(self):
        while self.spinning:
            sys.stdout.write(next(self.spinner_generator))
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write('\b')
            sys.stdout.flush()

    def start(self):
        self.spinning = True
        threading.Thread(target=self.run).start()

    def stop(self):
        self.spinning = False
        time.sleep(self.delay)



#!/usr/bin/env python

from watchme.watchers.psutils.decorators import monitor_resources
from time import sleep

# Here we create a decorator to monitor "my func." Specifically, we:
# - want to use the watcher "decorator" that already exists. If we want to
#   create on the fly, we would set creat=True
# - will record metrics every 3 seconds
# - to have somewhat of an impact on system resources we make a long list
# - we test to ensure that something is returned ("Hello!")

@monitor_resources('system', seconds=3)
def myfunc(iters, pause):
    long_list = []
    print("Generating a long list, pause is %s and iters is %s" % (pause, iters))
    for i in range(iters):
        long_list = long_list + (i*10)*['pancakes']
        print("i is %s, sleeping %s seconds" % (i, pause))
        sleep(pause)

    return len(long_list)

# ensure the function runs when the file is called
if __name__ == '__main__':
    print("Calling myfunc with 2 iters")
    result = myfunc(2, 2)
    print("Result list has length %s" % result)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from multiprocessing import (
    Process, 
    Queue
)
from functools import wraps
from watchme.logger import bot
from watchme.watchers.psutils import Task
from watchme import get_watcher
from time import sleep
import os


class ProcessRunner():

    def __init__(self, seconds=3, skip=[], include=[], only=[]):
        self.process = None
        self.seconds = seconds
        self.queue = Queue()
        self.timepoints = []

        # Export seconds to the environment
        os.environ["WATCHMEENV_SECONDS"] = str(self.seconds)

        # Ensure we have csv lists
        self.only = self._parse_custom(only)
        self.skip = self._parse_custom(skip)
        self.include = self._parse_custom(include)

    def _parse_custom(self, listy):
        '''parse an actual list (['one','two','three']) into 
           a csv list. If we don't have a list, ignore and assume already
           parsed that way.
 
           Parameters
           ==========
           listy: the actual list
        '''
        if isinstance(listy, list):
            listy = ','.join(listy)
        return listy

    @staticmethod
    def _wrapper(func, queue, args, kwargs):
        ret = func(*args, **kwargs)
        queue.put(ret)

    def run(self, func, *args, **kwargs):
        args2 = [func, self.queue, args, kwargs]
        p = Process(target=self._wrapper, args=args2)
        self.process = p
        p.start()

    def wait(self):

        # Parameters for the pid, and to skip sections of results
        params = {"skip": self.skip,
                  "pid": self.process.pid,
                  "include": self.include,
                  "only": self.only,
                  'func': 'monitor_pid_task'}

        # This particular decorator doesn't take input params
        task = Task("monitor_pid_task", params=params)

        # Export parameters and functions            
        function = task.export_func()
        params = task.export_params()

        # collect resources, then sleep
        while self.process.is_alive():

            # Function returns dictionary, we append to list of timepoints
            self.timepoints.append(function(**params))
            sleep(self.seconds)

        # Get the result, and the timepoints
        result = self.queue.get()
        self.process.join()
        return result


def monitor_resources(*args, **kwargs):
    '''a decorator to monitor a function every 3 (or user specified) seconds. 
       We include one or more task names that include data we want to extract.
       we get the pid of the running function, and then use the
       monitor_pid_task from psutils to watch it. The functools "wraps"
       ensures that the (fargs, fkwargs) are passed from the calling function
       despite the wrapper. The following parameters can be provided to
       "monitor resources"

       Parameters
       ==========
       watcher: the watcher instance to use, used to save data to a "task"
                folder that starts with "decorator-<name<"
       seconds: how often to collect data during the run.
       only: ignore skip and include, only include this custom subset
       skip: Fields in the result to skip (list).
       include: Fields in the result to include back in (list).
       create: whether to create the watcher on the fly (default False, must
               exist)
       name: the suffix of the decorator-psutils-<name> folder. If not provided,
             defaults to the function name
    '''
    def inner(func):

        @wraps(func)
        def wrapper(*fargs, **fkwargs):

            # Typically the task folder is the index, so we will create
            # indices that start with decorator-<task>
            result = None

            # The watcher is required, first keyword argument
            if len(args) == 0:
                bot.error("A watcher name is required for the psutils decorator.")
                return result

            # Get a watcher to save results to
            watcher = get_watcher(args[0], create=kwargs.get('create', False))

            # Start the function
            runner = ProcessRunner(seconds=kwargs.get('seconds', 3),
                                   skip=kwargs.get('skip', []),
                                   include=kwargs.get('include', []),
                                   only=kwargs.get('only', []))

            runner.run(func, *fargs, **fkwargs)
            result = runner.wait()

            # Save results (finishing runs) - key is folder created
            name = kwargs.get('name', func.__name__)
            key = 'decorator-psutils-%s' % name
            results = {key: runner.timepoints}
            watcher.finish_runs(results)
            
            # Return function result to the user
            return result
        return wrapper
    return inner



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import bot
from watchme.utils import ( 
    which, 
    get_user
)
from watchme.command import (
    get_commits,
    git_show,
    git_date
)
import os
import json

# Data Exports


def export_dict(self, task,
                      filename, 
                      name=None,
                      export_json=False,
                      from_commit=None,
                      to_commit=None,
                      base=None):
    '''Export a data frame of changes for a filename over time.

       Parameters
       ==========
       task: the task folder for the watcher to look in
       name: the name of the watcher, defaults to the client's
       base: the base of watchme to look for the task folder
       from_commit: the commit to start at
       to_commit: the commit to go to
       grep: the expression to match (not used if None)
       filename: the filename to filter to. Includes all files if not specified.
    '''
    if name == None:
        name = self.name

    if base == None:
        base = self.base

    # Quit early if the task isn't there
    if not self.has_task(task) and not task.startswith('decorator'):
        bot.exit('%s is not a valid task or decorator for %s' % (task, name))

    repo = os.path.join(base, self.name)
    if not os.path.exists(repo):
        bot.exit('%s does not exist.' % repo)

    filepath = os.path.join(base, self.name, task, filename)

    # Ensure that the filename exists in the repository
    if not os.path.exists(filepath):
        bot.exit('%s does not exist for watcher %s' %(filepath, name))

    # Now filepath must be relative to the repo
    filepath = os.path.join(task, filename)

    commits = get_commits(repo=repo,
                          from_commit=from_commit, 
                          to_commit=to_commit,
                          grep="ADD results %s" % task,
                          filename=filepath)

    # Keep lists of commits, dates, content    
    result = {'commits': [], 'dates': [], 'content': []}

    # Empty content (or other) returns None
    for commit in commits:
        content = git_show(repo=repo, commit=commit, filename=filepath)

        if export_json is True:
            content = json.loads(content)

        # If it's a list, add it to content
        if isinstance(content, list):
            result['content'] += content
        # Otherwise, append
        else:
            result['content'].append(content)

        result['dates'].append(git_date(repo=repo, commit=commit))
        result['commits'].append(commit)
    return result



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.utils import generate_temporary_file
from watchme.logger import bot
from .helpers import (
    get_params, 
    get_results, 
    get_headers,
    parse_success_response
)
from requests.exceptions import HTTPError
import os
import re
import tempfile
import requests


def get_task(url, **kwargs):
    '''a simple task to use requests to get a url. By default, we return
       the raw response.

       Parameters
       ==========

       REQUIRED:
           url: a url to return the page for

       OPTIONAL
           regex: a regular expression to search the text for (not used w/ json)
           save_as: return the result to save as json
    '''
    results = []
    paramsets = get_params(kwargs)
    headers = get_headers(kwargs)

    for params in paramsets:
        response = requests.get(url, params=params, headers=headers)

        if response.status_code == 200:

            # Parse the response per the user's request
            result = parse_success_response(response, kwargs)
            results.append(result)

    results = [x for x in results if x]
 
    if len(results) == 0:
        results = None

    return results


def post_task(url, **kwargs):
    '''a simple task to use requests to post to. By default, we return json.

       Parameters
       ==========

       REQUIRED:
           url: a url to post to
    '''
    results = []

    # The json params can vary, but headers do not
    jsonlist = get_params(kwargs, key='json_param_')
    headers = get_headers(kwargs)

    # Loop through lists of json and headers
    for params in jsonlist:

        # Get the post response and proceed if successful
        response = requests.post(url, json=params, headers=headers)
        if response.status_code == 200:

            # Parse the response per the user's request
            result = parse_success_response(response, kwargs)
            results.append(result)

        else:
            bot.error("%s: %s" %(response.status_code, response.reason))

    results = [x for x in results if x]

    # Return None if no results found
    if len(results) == 0:
        results = None

    return results


def download_task(url, **kwargs):
    '''a simple task to use requests to get a url. By default, we return
       the raw response.

       Parameters
       ==========

       REQUIRED:
           url: a url to download (stream)

       OPTIONAL:
           write_format: to change from default "w"
           disable_ssl_check: set to anything to not verify (not recommended)
    '''
    result = None

    # Update the user what we are doing
    bot.verbose("Downloading %s" % url)

    # Use the basename or the user set file_name to write to
    file_name = kwargs.get('file_name', os.path.basename(url))
    destination = os.path.join(tempfile.gettempdir(), file_name)    
    verify = True

    # Does the user want to disable ssl?
    if "disable_ssl_check" in kwargs:
        if kwargs['disable_ssl_check']:
            bot.warning('Verify of certificates disabled! ::TESTING USE ONLY::')
            verify = False

    # If the user doesn't want to write, but maybe write binary
    fmt = kwargs.get('write_format', 'wb')
    headers = get_headers(kwargs)

    # Does the url being requested exist?
    if requests.head(url, verify=verify, headers=headers).status_code in [200, 401]:

        # Stream the response
        response = requests.get(url, verify=verify, stream=True, headers=headers)

        # Invalid permissions
        if response.status_code == 401:
            return result

        # Successful, stream to result destination
        if response.status_code == 200:

            chunk_size = 1 << 20
            with open(destination, fmt) as filey:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    filey.write(chunk)

            result = destination

    return result


def get_url_selection(url, **kwargs):
    '''select some content from a page dynamically, using selenium.

       Parameters
       ==========
       kwargs: a dictionary of key, value pairs provided by the user
    '''
    
    results = None
    selector = kwargs.get('selection', None)
    headers = get_headers(kwargs)

    if selector == None:
        bot.error('You must define the selection (e.g., selection@.main')
        return results

    # Does the user want to get text?
    get_text = False
    if kwargs.get('get_text') != None:
        get_text = True

    # Are we searching for a regular expression in the result?
    regex = kwargs.get('regex')

    # Does the user want to get one or more attributes?
    attributes = kwargs.get('attributes', None)
    if attributes != None:
        attributes = attributes.split(',') 

    # User can pass a parameter like url_param_<name>
    # url_param_page=1,2,3,4,5,6,7,8,9
    paramsets = get_params(kwargs)

    # Each is a dictionary of values
    results = []
    for params in paramsets:
 
        # Get the page
        results += get_results(url=url,
                               selector=selector,
                               headers=headers,
                               attributes=attributes,
                               params=params,
                               get_text=get_text,
                               regex=regex)

    # No results
    if len(results) == 0:
        results = None

    return results



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher
from watchme.logger import bot

def main(args, extra):
    '''deactivate one or more watchers
    '''    
    # Doesn't work if watcher not provided
    watcher = args.watcher[0]
    watcher = get_watcher(watcher, base=args.base)
        
    # The user is deactivating the entire watcher
    if extra == None:
        watcher.deactivate()       
    else:
        for name in extra:
            watcher.deactivate(name)




'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''
from watchme.utils import ( 
    run_command, 
    write_file,
    get_tmpdir
)
from watchme.logger import bot
from watchme.defaults import WATCHME_BASE_DIR
import os
from datetime import datetime
import shutil
import time

def git_pwd(func):
    '''ensure that we are in the repo present working directory before running
       a git command. Return to where we were before after completion.
       The repo is always the first (positional or keyword) argument.
    '''      
    def git_pwd_inner(*args, **kwargs): 

        # Repo is either provided as a keyword argument, or the first positionl
        repo = kwargs.get('repo', '')

        # If provided as a positional argument
        if repo == '' and len(args) > 0:
            repo = args[0]

        # Keep a record of the present working directory
        pwd = os.getcwd()
        os.chdir(repo)
          
        # Run the git command         
        result = func(*args, **kwargs) 

        # Return to where we were before
        os.chdir(pwd)
        return result
    
    return git_pwd_inner


@git_pwd
def git_commit(repo, task, message):
    '''Commit to the git repo with a particular message. folder.

       Parameters
       ==========
       repo: the repository to commit to.
       task: the name of the task to add to the commit message
       message: the message for the commit, passed from the client
    '''
    # Commit with the watch group and date string
    message = 'watchme %s %s' %(task, message)

    # Commit
    command = 'git commit -a -m "%s"' % message
    bot.debug(command)
    run_command(command)

@git_pwd
def write_timestamp(repo, task, filename='TIMESTAMP'):
    '''write a file that includes the last run timestamp. This should be written
       in each task folder after a run.

       Parameters
       ==========
       repo: the repository to write the TIMESTAMP file to
       task: the name of the task folder to write the file to
       filename: the filename (defaults to TIMESTAMP)
    '''
    ts = time.time()
    strtime = datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
    filename = os.path.join(repo, task, filename)
    write_file(filename, strtime)
    return filename


@git_pwd
def git_date(repo, commit):
    '''get the date for a particular commit.
    
       Parameters
       ==========
       repo: the full path to the repository
       commit: the commit to get the date for
    '''
    command = 'git show -s --format=' + "%ci " + commit
    bot.debug(command)
    result = run_command(command)
    if result['return_code'] == 0:
        return result['message'].strip('\n')
    

@git_pwd
def git_show(repo, commit, filename):
    '''git show is used to pipe the content of a file at a particular commit
       to the screen (and calling python client). We must be in the $PWD of the
       repo for this to work.

       Parameters
       ==========
       repo: the repository to interact with
       commit: the commit to investigate for the file
       filename: the relative path to the file
    '''
    command = 'git show %s:%s' %(commit, filename)
    bot.debug(command)
    result = run_command(command)
    if result['return_code'] == 0:
        return result['message'].strip('\n')



def git_clone(repo, name=None, base=None, force=False):
    '''clone a git repo to a destination. The user can provide the following
       groupings of arguments:

       base without name: destination is ignored, the repo is cloned (named as
       it is) to the base. If the folder exists, --force must be used to remove
       it first.

       base with name: destination is ignored, repo is cloned (and named based
       on name variable) to the base. The same applies for force.

       dest provided: the repo is cloned to the destination, if it doesn't exist
       and/or force is True.

       Parameters
       ==========
       name: the name of the watcher to add
       base: the base of the watcher (defaults to $HOME/.watchme
       force: remove first if already exists
    '''
    if base == None:
        base = WATCHME_BASE_DIR

    # Derive the repository name
    if name == None:
        name = os.path.basename(repo).replace('.git', '')

    # First clone to temporary directory
    tmpdir = get_tmpdir()
    command = 'git clone %s %s' % (repo, tmpdir)    
    bot.debug(command)
    run_command(command)

    # ensure there is a watchme.cfg
    if not os.path.exists(os.path.join(tmpdir, 'watchme.cfg')):
        shutil.rmtree(tmpdir)
        bot.exit('No watchme.cfg found in %s, aborting.' % repo)

    # If it's good, move the repository
    dest = os.path.join(base, name)

    # Don't allow for overwrite
    if os.path.exists(dest): 
        if force is False:
            shutil.rmtree(tmpdir)        
            bot.exit('%s exists. Use --force to overwrite' % dest)
        else:
            shutil.rmtree(dest)

    # Move the repository there
    shutil.move(tmpdir, dest)

    # Ensure we don't sign gpg key
    run_command("git --git-dir=%s/.git config commit.gpgsign false" % dest)
    bot.info('Added watcher %s' % name)
    

@git_pwd
def get_commits(repo, from_commit=None, to_commit=None, grep=None, filename=None):
    '''get commits, starting from and going to a particular commit. if grep
       is defined, filter commits to those with messages that match that
       particular expression

       Parameters
       ==========
       from_commit: the commit to start at
       to_commit: the commit to go to
       grep: the expression to match (not used if None)
       filename: the filename to filter to. Includes all files if not specified.
    '''
    command = 'git log --all --oneline --pretty=tformat:"%H"' 

    # The earliest commit
    if from_commit == None:
        from_commit = get_earliest_commit()

    # The latest commit
    if to_commit == None:
        to_commit = get_latest_commit()

    # A regular expression to search for (and filter commits)
    if grep != None:
        command = "%s --grep \"ADD results\"" % command

    # Add the commit range
    command = "%s %s..%s" % (command, from_commit, to_commit)

    if filename != None:
        command = "%s -- %s" %(command, filename)

    bot.info(command)
    results = run_command(command)['message']
    results = [x for x in results.split('\n') if x]
    return results


def get_earliest_commit():
    '''get the earliest commit for a repository. This is intended to be used
       when in the present working directory

       Parameters
       ==========
       repo: the repository path to get the commit from
    '''
    commit = run_command('git rev-list --max-parents=0 HEAD')['message']
    return commit.strip('\n')


def get_latest_commit():
    '''get the latest commit for a repository in the present working directory

       Parameters
       ==========
       repo: the repository path to get the commit from
    '''
    commit = run_command('git log -n 1 --pretty=format:"%H"')['message']
    return commit.strip('\n')

@git_pwd
def git_add(repo, files):
    '''add one or more files to the git repo.

       Parameters
       ==========
       repo: the repository to commit to.
       files: one or more files to add.
    '''
    if not isinstance(files, list):
        files = [files]

    for f in files:
        command = 'git add %s' % f
        bot.debug(command)
        run_command(command)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import bot
import os


def remove_setting(self, section, name, save=True):
    '''remove a setting from the configuration file

       Parameters
       ==========
       section: the name of the section (task)
       name: the name of the variable to remove
       save: save the configuration file (default is True)
    '''
    removed = False
    self.load_config()

    # Remove the named section, if it exists
    if section in self.config:
        if name.lower() in self.config[section]:
            removed = self.config.remove_option(section, name)

    # Does the user want to save the file?
    if removed and save:
        self.save()

    return removed


def remove_section(self, section, save=True):
    '''remove a setting from the configuration file

       Parameters
       ==========
       section: the name of the section (task)
       save: save the configuration file (default is True)
    '''
    removed = False
    self.load_config()

    # Remove the named section, if it exists
    if section in self.config:
        self.config.remove_section(section)
        removed = True

    # Does the user want to save the file?
    if removed and save:
        self.save()

    return removed

def print_add_task(self, task):
    '''assemble a task section into a command that can create/add it.

       Parameters
       ==========
       task: the name of the task to inspect
    '''
    self.load_config()

    if task in self.config:
        command = "watchme add-task %s" % task
        for key in self.config[task]:
            value = self.config[task][key]
            command = "%s %s@%s" %(command, key, value)

        bot.info(command)
    else:
        bot.exit('%s is not a valid task.' % task)


def print_section(self, section):
    '''print a section (usually a task) from a configuration file,
       if it exists.

       Parameters
       ==========
       section: the name of the section (task)
    '''
    self.load_config()

    if section in self.config:
        bot.info('[%s]' % section)
        for key in self.config[section]:
            value = self.config[section][key]
            bot.custom(prefix=key, message=" = %s" % value, color="CYAN")
    else:
        bot.exit('%s is not a valid section.' % section)


def get_setting(self, section, name, default=None):
    '''return a setting from the config, if defined. Otherwise return
       default (None or set by user)
       
       Parameters
       ==========
       section: the section in the config, defaults to self.name
       name: they key (index) of the setting to look up
       default: (optional) if not found, return default instead.
    ''' 
    self.load_config()

    setting = None
    if section in self.config:
         if name.lower() in self.config[section]:
            setting = self.config[section][name.lower()]

    if setting is None and default is not None:
        setting = default

    return setting


def has_setting(self, section, name):
    '''return a boolean if a config has a setting (or not)
       Parameters
       ==========
       section: the section in the config, defaults to self.name
       name: they key (index) of the setting to look up
    ''' 
    self.load_config()

    exists = False
    if section in self.config:
         if name.lower() in self.config[section]:
            exists = True

    return exists


def has_section(self, section):
    '''return a boolean if a config has a section (e.g., a task or exporter)
       Parameters
       ==========
       section: the section in the config
    ''' 
    self.load_config()
    return section in self.config

def set_setting(self, section, key, value):
    '''set a key value pair in a section, if the section exists. Returns
       a boolean (True or False) to indicate if added.

       Parameters
       ==========
       section: the section in the config, defaults to self.name
       key: they key (index) of the setting to set
       value: the value to set.
    ''' 
    self.load_config()

    was_set = False
    if section in self.config:
        self.config[section][key] = value
        was_set = True

    return was_set


def get_section(self, name):
    '''get a section from the config, if it exists
    '''
    section = None
    self.load_config()
    if name in self.config:
        section = self.config[name]
    return section



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import bot
from watchme.defaults import WATCHME_WORKERS
import multiprocessing
import itertools
import time
import signal
import sys
import re
import os


class Workers(object):

    def __init__(self, workers=None, show_progress=False):

        if workers is None:
            workers = WATCHME_WORKERS
        self.workers = workers
        self.show_progress = show_progress
        bot.debug("Using %s workers for multiprocess." % (self.workers))

    def start(self):
        bot.debug("Starting multiprocess")
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()
        self.runtime = self.runtime = self.end_time - self.start_time
        bot.debug("Ending multiprocess, runtime: %s sec" % (self.runtime))

    def run(self, funcs, tasks):
        '''run will send a list of tasks, a tuple with arguments, through a function.
           the arguments should be ordered correctly.
        
           Parameters
           ==========
           funcs: the functions to run with multiprocessing.pool, a dictionary
                  with lookup by the task name
           tasks: a dict of tasks, each task name (key) with a 
                  tuple of arguments to process
        '''
        # Number of tasks must == number of functions
        assert(len(funcs)==len(tasks))

        # Keep track of some progress for the user
        progress = 1
        total = len(tasks)

        # if we don't have tasks, don't run
        if len(tasks) == 0:
            return

        # results will also have the same key to look up
        finished = dict()
        results = []

        try:
            prefix = "[%s/%s]" % (progress, total)
            if self.show_progress:
                bot.show_progress(0, total, length=35, prefix=prefix)
            pool = multiprocessing.Pool(self.workers, init_worker)

            self.start()
            for key, params in tasks.items():
                func = funcs[key]
                if not self.show_progress:
                    bot.info('Processing task %s:%s' % (key, params))
                result = pool.apply_async(multi_wrapper,
                                          multi_package(func, [params]))
                
                # Store the key with the result
                results.append((key, result))


            while len(results) > 0:
                pair = results.pop()
                key, result = pair
                result.wait()
                if self.show_progress:
                    bot.show_progress(progress, total, length=35, prefix=prefix)
                progress += 1
                prefix = "[%s/%s]" % (progress, total)
                finished[key] = result.get()

            self.end()
            pool.close()
            pool.join()

        except (KeyboardInterrupt, SystemExit):
            bot.error("Keyboard interrupt detected, terminating workers!")
            pool.terminate()
            sys.exit(1)

        except Exception as e:
            bot.error(e)

        return finished


# Supporting functions for MultiProcess Worker
def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def multi_wrapper(func_args):
    function, kwargs = func_args
    return function(**kwargs)

def multi_package(func, kwargs):
    zipped = zip(itertools.repeat(func), kwargs)
    return zipped



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

The watcher is actually a connection to crontab. This is what helps to schedule
the watched to check for changes at some frequency, and update the files.


'''

from watchme.utils import get_userhome
import tempfile
import os
import sys


################################################################################
# environment / options
################################################################################


def getenv(variable_key, default=None, required=False, silent=True):
    ''' attempt to get an environment variable. If the variable
        is not found, None is returned.

        Parameters
        ==========
        variable_key: the variable name
        required: exit with error if not found
        silent: Do not print debugging information for variable
    '''
    variable = os.environ.get(variable_key, default)
    if variable is None and required:
        bot.error("Cannot find environment variable %s, exiting." %variable_key)
        sys.exit(1)

    if not silent and variable is not None:
            bot.verbose("%s found as %s" %(variable_key,variable))

    return variable



################################################################################
# Helpme

USERHOME = get_userhome()

WATCHME_WATCHER = getenv('WATCHME_WATCHER', 'watcher')
_config = os.path.join(USERHOME, ".watchme")
WATCHME_BASE_DIR = getenv('WATCHME_BASE_DIR', _config)
WATCHME_WORKERS = int(getenv('WATCHME_WORKERS', 9))

# The types of valid watchers (currently only urls). Corresponds with
# a folder under "main/watchers"
WATCHME_TASK_TYPES = ['urls', 'url', 'psutils', 'results']
WATCHME_DEFAULT_TYPE = "urls"

# Parameters not allowed for task clients, set by TaskBase
WATCHME_NOTALLOWED_PARAMS = ['type', 'active']



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher
from watchme.logger import bot

def main(args, extra):
    '''activate one or more watchers
    '''    
    # Required - will print help if not provided
    name = args.watcher[0]

    # Exit if the user doesn't provide a time
    if extra == None:
        bot.exit('Please provide a time frame (@daily, @hourly, @weekly, etc.)')
              
    # Determine the time to use
    if '@daily' in extra:
        minute, hour, month, day, weekday = 0, 0, '*', '*', '*'
    elif "@hourly" in extra:
        minute, hour, month, day, weekday = 0, '*', '*', '*', '*'
    elif "@weekly" in extra:
        minute, hour, month, day, weekday = 0, 0, '*', '*', 0
    elif "@monthly" in extra:
        minute, hour, month, day, weekday = 0, 0, 1, '*', '*'
    elif "@yearly" in extra:
        minute, hour, month, day, weekday = 0, 0, 1, 1, '*'
    else:
        if len(extra) != 5:
            message = '''Please enter a frequency (@weekly) or use a valid 
                         cron timestamp (see https://crontab.guru/).'''
            bot.exit(message)
        minute, hour, month, day, weekday = extra

    # Schedule the watcher
    watcher = get_watcher(name, base=args.base) 
    watcher.schedule(minute, hour, month, day, weekday, force=args.force)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


'''

from .create import (
    create_watcher_base,
    create_watcher
)

from .commit import (
    write_timestamp,
    get_earliest_commit,
    get_commits,
    git_commit, 
    git_clone, 
    git_add,
    git_date,
    git_show
)
from .utils import (
    get_watchers,
    list_task,
    list_watcher,
    list_watcher_types
)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

import os
import tempfile
import requests
import re

# Helper functions

def get_params(kwargs, key='url_param_'):
    '''a general function to get parameter sets based on a user input. 
       Returns a list of dictionaries, one per set.

       Parameters
       ==========
       kwargs: the dictionary of keyword arguments that may contain url
               parameters (format is url_param_<name>
       key: the string that the parameters start with (defaults to url_param)
    '''
    # Keey dictionary based on index of param
    params = {}

    names = [x for x in kwargs if x.startswith(key)]
    for n in range(len(names)):
        name = names[n]

        # Params are split by commas, with index corresponding to list index
        paramlist = kwargs.get(name).split(',')

        # Remove the "url_param"
        name = name.replace(key, '', 1)

        # Update the dictionary of dictionaries
        for i in range(len(paramlist)):

            # Add the index to the dicionary (will be turned in list later)
            if i not in params:
                params[i] = {}

            if paramlist[i] != '':
                params[i][name] = paramlist[i]

    # Unwrap the list
    params = [x for x in params.values()]

    # If no special params, we at least need to run once
    if len(params) == 0:
        params = [{}]

    return params


def parse_success_response(response, kwargs):
    '''parse a successful response of 200, meaning we honor the user
       request to return json, search for a regular expression, or return
       raw text. This is used by the basic GET/POST functions. For parsing
       with beautiful soup, see "get_results" and "get_url_selection"

       Parameters
       ==========
       response: the requests (200) response
       kwargs: dictionary of keyword arguments provided to function
    '''
    result = None
    save_as = kwargs.get('save_as', 'json')
    regex = kwargs.get('regex')

    # Returning the result as json will detect dictionary, and save json
    if save_as == "json":
        result = response.json()

    # As an alternative, search for a regular expression
    elif regex not in ["", None]:
        match = re.search(regex, response.text)
        result = match.group()

    # Otherwise, we return text
    else:
        result = response.text
    return result


def get_headers(kwargs):
    '''Get a single set of headers from the kwargs dict. A user agent is added
       as it is helpful in most cases.

       Parameters
       ==========
       kwargs: the dictionary of keyword arguments that may contain url
               parameters (format is url_param_<name>
    '''
    headers = {"User-Agent": "Mozilla/5.0"}

    for key, value in kwargs.items():
        if key.startswith('header_'):
            name = key.replace('header_', '', 1)

            # The header is defined with a value
            if value != None:
                headers[name] = value

            # If the user wants to remove the User-Agent (or any) header
            elif value == None and name in headers:
                del headers[name]

    return headers


def get_results(url, 
                selector,
                func=None,
                attributes=None,
                params={},
                get_text=False,
                headers={},
                regex=None):

    '''given a url, a function, an optional selector, optional attributes, 
       and a set (dict) of parameters, perform a request. This function is
       used if the calling function needs special parsing of the html with
       beautiful soup. If only a post/get is needed, this is not necessary.

       Parameters
       ==========
       url: the url to get (required)
       func: the function to use, defaults to requests.get
       selector:  selection for the html response
       attributes: optional, a list of attributes
       params: a dictionary of parameters
       headers: a dictionary of header key value pairs
    '''
    from bs4 import BeautifulSoup

    if func == None:
        func = requests.get

    response = func(url, params=params, headers=headers)
    results = []

    if response.status_code == 200:   
        soup = BeautifulSoup(response.text, 'lxml')

        # Get the selection
        for entry in soup.select(selector):

            # Does the user want to get attributes
            if attributes != None:
                [results.append(entry.get(x)) for x in attributes]

            # Second priority for regular expression on text
            elif regex not in [None, ""]:
                match = re.search(regex, entry.text)
                results.append(match.group())

            # Does the user want to get text?
            elif get_text == True:
                results.append(entry.text)

            # Otherwise, return the entire thing
            else:
                results.append(str(entry))

    # Clean up results
    results = [x for x in results if x]

    return results



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


'''

from watchme.version import __version__

def get_watcher(name="watcher", 
                base=None,
                create=False,
                quiet=True, **kwargs):
    '''
       get the correct watcher depending on the environment variable
       WATCHME_WATCHER or default to "watcher"

       Parameters
       ==========
       name: the name of the watcher, will be made all lowercase
       base: the watcher base, if not defined, will use WATCHER_BASE_DIR envar
       create: if the watcher folder doesn't exist, create it (default False)
               for all interactions with a watcher other than create, we should
               exit if the watcher the user wants doesn't exist.
       quiet: if True, suppress most output about the client (e.g. speak)
    '''
    from watchme.watchers import Watcher

    Watcher.name = name.lower()
    Watcher.quiet = quiet

    # Initialize the watcher
    return Watcher(create=create, base=base)
